/* $Id: multisync.c,v 1.8 1998/11/24 18:34:02 marcus Exp $
******************************************************************************

   Monitor driver for multisync monitors

   Copyright (C) 1995 Steffen Seeger	[seeger@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define DRIVER_NAME	"Multisync monitor driver"
#define DRIVER_REV	"$Revision: 1.8 $"

#ifdef DEBUG
#define NEED_MONITOR_SYMS
#endif

#include <kgi/module.h>
#include "monitors.inc"

static int param_vmin;
static int param_vmax;
static int param_hmin;
static int param_hmax;
#if 0
static int param_cmin;
#endif
static int param_cmax;

static kgi_subparam param_table[] = {
	{KGIMON_VMIN, "vmin", &param_vmin, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_VMAX, "vmax", &param_vmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
    	{KGIMON_HMIN, "hmin", &param_hmin, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{KGIMON_HMAX, "hmax", &param_hmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
#if 0 /* This shouln't be != 0 for any monitor... */
	{KGIMON_CMIN, "cmin", &param_cmin, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
#endif
	{KGIMON_CMAX, "bandwidth", &param_cmax, sizeof(int),
	 KGI_SUBPARAM_INT, 0644},
	{0}
};

static inline void update_params(void)
{
	param_vmin = monitor.vfreq.min;
	param_vmax = monitor.vfreq.max;
	param_hmin = monitor.hfreq.min;
	param_hmax = monitor.hfreq.max;
#if 0
	param_cmin = monitor.dclk.min;
#endif
	param_cmax = monitor.dclk.max;
}
static inline void update_monitor(void)
{
	monitor.vfreq.min = param_vmin;
	monitor.vfreq.max = param_vmax;
	monitor.hfreq.min = param_hmin;
	monitor.hfreq.max = param_hmax;
#if 0
	monitor.dclk.min = param_cmin;
#endif
	monitor.dclk.max = param_cmax;
}

int kgim_monitor_init(struct kgi_card *card, struct kgi_display *dpy)
{
	if (!(monitor.synctype & SYNC_MULTISYNC))
	{
		NOTICE0("The monitor configured is not of type MULTISYNC!");
		return -E(MONITOR,UNKNOWN);
	}

	update_params();

	DEBUG1("%s %s Monitor Driver",
	       monitor.manufact, monitor.model);
	DEBUG1("(%dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)",
	       monitor.maxdots.x, monitor.maxdots.y,
	       monitor.size.x, monitor.size.y,
	       monitor.hfreq.min / 1000 ,monitor.hfreq.max / 1000,
	       monitor.vfreq.min, monitor.vfreq.max);

	card->monitor = &monitor;
	return EOK;
}

void kgim_monitor_done(struct kgi_display *dpy)
{
	DEBUG1("%s %s Monitor driver removed.", 
	       monitor.manufact, monitor.model);
}

int kgim_monitor_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{ 
	int hfreq, vfreq, result;
	int width = mode->x.width;

	update_monitor();

	switch (*cmd)
	{
	case CMD_PROPOSE:
		mode->dclk = monitor.dclk.max;

		for(hfreq = monitor.hfreq.max; hfreq > monitor.hfreq.min;
		    mode->dclk -= 1000)
		{
			/*mode->x.syncstart = width * (width + 41360) / 40000;
			mode->x.syncend = width * (width * 7 / 10 + 8952) / 8000;
			mode->x.total = width * (width + 9360) / 8000;*/
			mode->x.syncstart = width * (1000 + 50) / 1000;
			mode->x.syncend = width * (1000 + 180) / 1000;
			mode->x.total = width * (1000 + 350) / 1000;

			/* !!! A little hack for the Mystique */
			if (!strcmp(dpy->card->model,"Mystique") &&
			    mode->request.graphtype == KGIGT_8BIT) {
				mode->x.total     = (mode->x.total     + 15) & ~15;
				mode->x.syncstart = (mode->x.syncstart + 15) & ~15;
				mode->x.syncend   = (mode->x.syncend   + 15) & ~15;
			}

			mode->x.blankstart = mode->x.syncstart;
			mode->x.blankend = mode->x.syncend;

			hfreq = mode->dclk / mode->x.total;
			if (monitor.hfreq.max < hfreq)
				continue;

			mode->y.syncstart = mode->y.width + 3;
			mode->y.syncend   = mode->y.width + 9;
			mode->y.total     = mode->y.width + hfreq / 1000 + 1; /* 1ms flyback */
			mode->y.blankstart= mode->y.syncstart; 
			mode->y.blankend  = mode->y.syncend;

			mode->x.polarity  = mode->y.polarity = 1; /* Doesn't matter */

			vfreq = mode->dclk / (mode->x.total * mode->y.total);

			if ((monitor.vfreq.min <= vfreq) && 
			    (vfreq <= monitor.vfreq.max))
			{
				
				if (monitor.flags & FL_MON_PROPSIZE)
				{
					mode->request.size.x = 
						mode->x.width * 
						monitor.size.x /
						monitor.maxdots.x;
					mode->request.size.y = 
						mode->y.width * 
						monitor.size.y /
						monitor.maxdots.y;
				}
				else 
				{
					
					mode->request.size = monitor.size;
				}

				DEBUG1("Calculated %i Hz refresh (%i Hz, "
				       "DCLK = %i Hz)",
				       vfreq, hfreq, mode->dclk);

				*cmd = CMD_LOWER;
				return EOK;
			}
		}
		
		NOTICE1("Invalid mode request.");
		return -E(MONITOR,UNKNOWN);
		break;

	case CMD_LOWER:
		if (mode->dclk < monitor.dclk.min)
		{
			
			return -E(MONITOR,UNKNOWN);
		}

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_RAISE:
		if (mode->dclk > monitor.dclk.max)
		{
			
			return -E(MONITOR,UNKNOWN);
		}
		

		*cmd = CMD_CHECK;
		return EOK;

	case CMD_CHECK: 
		hfreq = mode->dclk / mode->x.total;
		vfreq = mode->dclk / (mode->x.total * mode->y.total);
		result =
			(monitor.hfreq.min - (monitor.hfreq.min >> 7) <= hfreq) &&
	/*		(monitor.hfreq.min <= hfreq) && */
			(hfreq <= monitor.hfreq.max) && 
			(monitor.vfreq.min <= vfreq) && 
			(vfreq <= monitor.vfreq.max) && 
			(monitor.dclk.min <= mode->dclk) && 
			(mode->dclk <= monitor.dclk.max);
		if (result) 
		{
			
			DEBUG2("Accepted %i Hz refresh (%i Hz, DCLK=%i Hz)",
			       vfreq, hfreq, mode->dclk);
		}
		else 
		{
			
			DEBUG2("Refresh check failed. (%i Hz, %i Hz, %i Hz)",
			       vfreq, hfreq, mode->dclk);
		}
		
		*cmd = CMD_READY;
		return result ? EOK : -E(MONITOR,UNKNOWN);
		break;

	default:
		INTERNAL_ERROR;
		return -E(MONITOR,UNKNOWN);
	}
}

void kgim_monitor_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
}

int kgim_monitor_command(struct kgi_graphic *gr, kgi_uint cmd, kgi_uint arg)
{
	update_monitor();

	switch(cmd)
	{
	case MONITOR_GETINFO:
		gr->io_res = &monitor;
		return EOK;
		
	case MONITOR_SETINFO:
		if (suser())
		{
			
			memcpy(&monitor, gr->io_buf, sizeof(monitor));
			update_params();
			return EOK;
		}
		else
		{
			
			return -EPERM;
		}
		
	default:
		return -E(MONITOR,UNKNOWN);
	}
}

struct kgi_subparam *kgim_monitor_get_params(void)
{
	return param_table;
}
